// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2010-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Client.Session;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Pop3.Client {
  public partial class PopClient : IDisposable {
    public PopClientProfile Profile {
      get { return profile; }
    }

    public int Timeout {
      get { return (session == null) ? profile.Timeout : session.TransactionTimeout; }
      set
      {
        profile.Timeout = value;

        if (session != null)
          session.TransactionTimeout = value;
      }
    }

    public int SendTimeout {
      get { return (session == null) ? profile.SendTimeout : session.SendTimeout; }
      set
      {
        profile.SendTimeout = value;

        if (session != null)
          session.SendTimeout = value;
      }
    }

    public int ReceiveTimeout {
      get { return (session == null) ? profile.ReceiveTimeout : session.ReceiveTimeout; }
      set
      {
        profile.ReceiveTimeout = value;

        if (session != null)
          session.ReceiveTimeout = value;
      }
    }

    public bool IsConnected {
      get { return (session != null && session.State != PopSessionState.NotConnected); }
    }

    public bool IsBusy {
      get { return sessionStatus == SessionStatus_Busy; }
    }

    internal PopSession Session {
      get { ThrowIfNotConnected(); return session; }
    }

    public bool IsSecureSession {
      get { ThrowIfNotConnected(); return session.IsSecureConnection; }
    }

    public PopCapabilitySet ServerCapabilities {
      get { ThrowIfNotConnected(); return session.ServerCapabilities; }
    }

    /*
     * construction
     */
    public PopClient()
      : this(new PopClientProfile())
    {
    }

    public PopClient(Uri authority)
      : this(new PopClientProfile(authority))
    {
    }

    public PopClient(string host)
      : this(new PopClientProfile(host))
    {
    }

    public PopClient(string host, string userName)
      : this(new PopClientProfile(host, userName))
    {
    }

    public PopClient(string host, bool securePort, string userName)
      : this(new PopClientProfile(host, securePort, userName))
    {
    }

    public PopClient(string host, int port)
      : this(new PopClientProfile(host, port))
    {
    }

    public PopClient(string host, int port, string userName)
      : this(new PopClientProfile(host, port, userName))
    {
    }

    public PopClient(string host, int port, bool securePort)
      : this(new PopClientProfile(host, port, securePort))
    {
    }

    public PopClient(string host, int port, bool securePort, string userName)
      : this(new PopClientProfile(host, port, securePort, userName))
    {
    }

    public PopClient(string host, int port, string userName, string authType)
      : this(new PopClientProfile(host, port, userName, authType))
    {
    }

    public PopClient(string host, int port, bool securePort, string userName, string authType)
      : this(new PopClientProfile(host, port, securePort, userName, authType))
    {
    }

    public PopClient(string host, int port, bool securePort, string userName, string authType, int timeout)
      : this(new PopClientProfile(host, port, securePort, userName, authType, timeout))
    {
    }

    public PopClient(PopClientProfile profile)
    {
      if (profile == null)
        throw new ArgumentNullException(nameof(profile));

      this.profile = profile;
    }

    /*
     * connect impl
     */
    private static PopSession CreateSession(ConnectParams @params)
    {
      try {
        return PopSessionCreator.CreateSession(@params.Profile,
                                               @params.AuthMechanism,
                                               @params.CreateSslStreamCallback ?? PopSslConnection.CreateSslStream,
                                               @params.CancellationToken);
      }
      finally {
        @params.Profile.SetCredentials(null);
      }
    }

    internal class ConnectParams {
      public readonly PopClientProfile Profile;
      public readonly SaslClientMechanism AuthMechanism;
      public readonly UpgradeConnectionStreamCallback CreateSslStreamCallback;
      public readonly CancellationToken CancellationToken;

      public ConnectParams(PopClientProfile profile,
                           ICredentialsByHost credentials,
                           UpgradeConnectionStreamCallback createSslStreamCallback,
                           CancellationToken cancellationToken)
      {
        Profile = profile.Clone();
        Profile.SetCredentials(credentials);

        CreateSslStreamCallback = createSslStreamCallback;
        CancellationToken = cancellationToken;
      }

      public ConnectParams(PopClientProfile profile,
                           SaslClientMechanism authMechanism,
                           UpgradeConnectionStreamCallback createSslStreamCallback,
                           CancellationToken cancellationToken)
      {
        Profile = profile.Clone();
        AuthMechanism = authMechanism;
        CreateSslStreamCallback = createSslStreamCallback;
        CancellationToken = cancellationToken;
      }
    }

    /*
     * BeginConnect()/EndConnect()
     */
    public IAsyncResult BeginConnect(string password)
    {
      return BeginConnect(password, null, null, null);
    }

    public IAsyncResult BeginConnect(string password,
                                     UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return BeginConnect(password, createSslStreamCallback, null, null);
    }

    public IAsyncResult BeginConnect(string password,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(password, null, asyncCallback, asyncState);
    }

    public IAsyncResult BeginConnect(string password,
                                     UpgradeConnectionStreamCallback createSslStreamCallback,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(new ConnectParams(profile,
                                            new NetworkCredential(profile.UserName, password),
                                            createSslStreamCallback,
                                            CancellationToken.None),
                          asyncCallback,
                          asyncState);
    }

    public IAsyncResult BeginConnect(ICredentialsByHost credentials)
    {
      return BeginConnect(credentials, null, null, null);
    }

    public IAsyncResult BeginConnect(ICredentialsByHost credentials,
                                     UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return BeginConnect(credentials, createSslStreamCallback, null, null);
    }

    public IAsyncResult BeginConnect(ICredentialsByHost credentials,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(credentials, null, asyncCallback, asyncState);
    }

    public IAsyncResult BeginConnect(ICredentialsByHost credentials,
                                     UpgradeConnectionStreamCallback createSslStreamCallback,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(new ConnectParams(profile,
                                            credentials,
                                            createSslStreamCallback,
                                            CancellationToken.None),
                          asyncCallback,
                          asyncState);
    }

    public IAsyncResult BeginConnect(SaslClientMechanism authMechanism)
    {
      return BeginConnect(authMechanism, null, null, null);
    }

    public IAsyncResult BeginConnect(SaslClientMechanism authMechanism,
                                     UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return BeginConnect(authMechanism, createSslStreamCallback, null, null);
    }

    public IAsyncResult BeginConnect(SaslClientMechanism authMechanism,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(authMechanism, null, asyncCallback, asyncState);
    }

    public IAsyncResult BeginConnect(SaslClientMechanism authMechanism,
                                     UpgradeConnectionStreamCallback createSslStreamCallback,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(new ConnectParams(profile,
                                            authMechanism,
                                            createSslStreamCallback,
                                            CancellationToken.None),
                          asyncCallback,
                          asyncState);
    }

    private Task beginEndConnectTask = null;

    private IAsyncResult BeginConnect(ConnectParams @params,
                                      AsyncCallback asyncCallback,
                                      object asyncState)
    {
      var task = ConnectAsync(@params);
      var tcs = new TaskCompletionSource<object>(asyncState);

      task.ContinueWith(t => {
        if (t.IsFaulted)
          tcs.TrySetException(t.Exception.InnerExceptions);
        else if (t.IsCanceled)
          tcs.TrySetCanceled();
        else
          tcs.TrySetResult(null);

        asyncCallback?.Invoke(tcs.Task);
      }, TaskScheduler.Default);

      beginEndConnectTask = tcs.Task;

      return beginEndConnectTask;
    }

    public void EndConnect(IAsyncResult asyncResult)
    {
      if (asyncResult == null)
        throw new ArgumentNullException(nameof(asyncResult));
      if (asyncResult != beginEndConnectTask)
        throw ExceptionUtils.CreateArgumentMustBeValidIAsyncResult(nameof(asyncResult));

      try {
        beginEndConnectTask.Wait();
      }
      finally {
        beginEndConnectTask = null;
      }
    }

    /*
     * ConnectAsync()
     */
    public Task ConnectAsync(string password)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            new NetworkCredential(profile.UserName, password),
                                            null,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(string password,
                             UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            new NetworkCredential(profile.UserName, password),
                                            createSslStreamCallback,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(string password,
                             UpgradeConnectionStreamCallback createSslStreamCallback,
                             CancellationToken cancellationToken)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            new NetworkCredential(profile.UserName, password),
                                            createSslStreamCallback,
                                            cancellationToken));
    }

    public Task ConnectAsync(ICredentialsByHost credentials)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            credentials,
                                            null,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(ICredentialsByHost credentials,
                             UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            credentials,
                                            createSslStreamCallback,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(ICredentialsByHost credentials,
                             UpgradeConnectionStreamCallback createSslStreamCallback,
                             CancellationToken cancellationToken)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            credentials,
                                            createSslStreamCallback,
                                            cancellationToken));
    }

    public Task ConnectAsync(SaslClientMechanism authMechanism)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            authMechanism,
                                            null,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(SaslClientMechanism authMechanism,
                             UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            authMechanism,
                                            createSslStreamCallback,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(SaslClientMechanism authMechanism,
                             UpgradeConnectionStreamCallback createSslStreamCallback,
                             CancellationToken cancellationToken)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            authMechanism,
                                            createSslStreamCallback,
                                            cancellationToken));
    }

    private Task ConnectAsync(ConnectParams @params)
    {
      ThrowIfAlreadyConnected();

      byDisposeCancellationTokenSource?.Dispose(); // dispose previously used CancellationTokenSource

      byDisposeCancellationTokenSource = new CancellationTokenSource();

      return StartTaskExclusively(() => {
        PopSession createdSession = null;

        try {
          createdSession = PopClient.CreateSession(@params);
        }
        finally {
          if (byDisposeCancellationTokenSource.IsCancellationRequested)
            createdSession?.Disconnect(false); // ignore exceptions
          else
            this.session = createdSession;
        }
      }, @params.CancellationToken, false);
    }

    /*
     * Connect()
     */
    public void Connect(string password)
    {
      Connect(new ConnectParams(profile,
                                new NetworkCredential(profile.UserName, password),
                                null,
                                CancellationToken.None));
    }

    public void Connect(string password,
                        UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      Connect(new ConnectParams(profile,
                                new NetworkCredential(profile.UserName, password),
                                createSslStreamCallback,
                                CancellationToken.None));
    }

    public void Connect(string password,
                        UpgradeConnectionStreamCallback createSslStreamCallback,
                        CancellationToken cancellationToken)
    {
      Connect(new ConnectParams(profile,
                                new NetworkCredential(profile.UserName, password),
                                createSslStreamCallback,
                                cancellationToken));
    }

    public void Connect(ICredentialsByHost credentials)
    {
      Connect(new ConnectParams(profile,
                                credentials,
                                null,
                                CancellationToken.None));
    }

    public void Connect(ICredentialsByHost credentials,
                        UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      Connect(new ConnectParams(profile,
                                credentials,
                                createSslStreamCallback,
                                CancellationToken.None));
    }

    public void Connect(ICredentialsByHost credentials,
                        UpgradeConnectionStreamCallback createSslStreamCallback,
                        CancellationToken cancellationToken)
    {
      Connect(new ConnectParams(profile,
                                credentials,
                                createSslStreamCallback,
                                cancellationToken));
    }

    public void Connect(SaslClientMechanism authMechanism)
    {
      if (authMechanism == null)
        throw new ArgumentNullException(nameof(authMechanism));

      Connect(new ConnectParams(profile,
                                authMechanism,
                                null,
                                CancellationToken.None));
    }

    public void Connect(SaslClientMechanism authMechanism,
                        UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      if (authMechanism == null)
        throw new ArgumentNullException(nameof(authMechanism));

      Connect(new ConnectParams(profile,
                                authMechanism,
                                createSslStreamCallback,
                                CancellationToken.None));
    }

    public void Connect(SaslClientMechanism authMechanism,
                        UpgradeConnectionStreamCallback createSslStreamCallback,
                        CancellationToken cancellationToken)
    {
      if (authMechanism == null)
        throw new ArgumentNullException(nameof(authMechanism));

      Connect(new ConnectParams(profile,
                                authMechanism,
                                createSslStreamCallback,
                                cancellationToken));
    }

    private void Connect(ConnectParams @params)
    {
      ThrowIfAlreadyConnected();

      RunExclusively(() => {
        session = CreateSession(@params);
      }, false);
    }

    /*
     * destruction
     */
    void IDisposable.Dispose()
    {
      DetachFromSession();

      if (session != null) {
        (session as IDisposable).Dispose();
        session = null;
      }

      byDisposeCancellationTokenSource?.Cancel(); // cancel and dispose async-connected session if created
    }

    private void DetachFromSession()
    {
      lock (messages) {
        foreach (var message in messages.Values) {
          message.Client = null;
        }

        messages.Clear();
      }
    }

    public void Disconnect()
    {
      Disconnect(false);
    }

    public void Logout()
    {
      Disconnect(true);
    }

    public void Disconnect(bool logout)
    {
      RunExclusively(() => {
        if (session == null)
          return;

        DetachFromSession();

        session.Disconnect(logout);
        session = null;
      }, false);
    }

    /*
     * methods for exclusive execution
     */
    private const int SessionStatus_Ready = 0;
    private const int SessionStatus_Busy = 1;

    private int sessionStatus = SessionStatus_Ready;

    private void SetBusy()
    {
      if (SessionStatus_Busy == Interlocked.CompareExchange(ref sessionStatus, SessionStatus_Busy, SessionStatus_Ready))
        throw new InvalidOperationException("client is busy; concurrent operations are not supported");
    }

    private void UnsetBusy()
    {
      Interlocked.Exchange(ref sessionStatus, SessionStatus_Ready);
    }

    internal void RunExclusively(Action operation, bool throwIfNotConnected = true)
    {
      if (throwIfNotConnected)
        ThrowIfNotConnected();

      SetBusy();

      try {
        operation();
      }
      finally {
        UnsetBusy();
      }
    }

    internal T RunExclusively<T>(Func<T> operation)
    {
      ThrowIfNotConnected();

      SetBusy();

      try {
        return operation();
      }
      finally {
        UnsetBusy();
      }
    }

    internal Task StartTaskExclusively(Action operation,
                                       CancellationToken cancellationToken,
                                       bool throwIfNotConnected = true)
    {
      if (throwIfNotConnected)
        ThrowIfNotConnected();

      SetBusy();

      var t = Task.Factory.StartNew(() => {
        try {
          operation();
        }
        finally {
          UnsetBusy();
        }
      }, cancellationToken);

      // XXX
      if (cancellationToken.IsCancellationRequested && t.Status == TaskStatus.WaitingToRun)
        UnsetBusy();
      else if (t.IsCanceled)
        UnsetBusy();

      return t;
    }

    internal Task<T> StartTaskExclusively<T>(Func<T> operation,
                                             CancellationToken cancellationToken)
    {
      ThrowIfNotConnected();

      SetBusy();

      var t = Task.Factory.StartNew(() => {
        try {
          return operation();
        }
        finally {
          UnsetBusy();
        }
      }, cancellationToken);

      // XXX
      if (cancellationToken.IsCancellationRequested && t.Status == TaskStatus.WaitingToRun)
        UnsetBusy();
      else if (t.IsCanceled)
        UnsetBusy();

      return t;
    }

    /*
     * methods for checking and validating internal status
     */
    private void ThrowIfAlreadyConnected()
    {
      if (IsConnected)
        throw new InvalidOperationException("already connected");
    }

    private void ThrowIfNotConnected()
    {
      if (!IsConnected)
        throw new InvalidOperationException("not connected or already disconnected");
    }

#if DEBUG
    internal void ThrowIfNotInExclusiveOperation()
    {
      if (sessionStatus == SessionStatus_Ready)
        throw new InvalidOperationException("not in exclusive operation");
    }
#endif

    public override string ToString()
    {
      return string.Format("{{PopClient: Authority='{0}', IsConnected={1}}}",
                           profile.Authority,
                           IsConnected);
    }

    private PopSession session;
    private PopClientProfile profile;
    private CancellationTokenSource byDisposeCancellationTokenSource;
  }
}
