using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Sockets;
using System.Threading;

using Smdn.Text.Encodings;

namespace Smdn.Net.Pop3.Client {
  public class PopPseudoServer : PseudoServer {
    private class Response : ResponseBase {
      public bool IsQuitResponse {
        get; set;
      }

      public bool IsPushResponse {
        get; set;
      }

      public Response(string expectedCommand,
                      string responseText,
                      Action actionBeforeRespond,
                      Action actionAfterRespond)
        : base(expectedCommand,
               responseText,
               actionBeforeRespond,
               actionAfterRespond)
      {
      }
    }

    public PopPseudoServer()
      : base()
    {
    }

    protected override void Initialize()
    {
      base.Initialize();

      SetGreetingResponse("+OK\r\n");
    }

    public static void Start(Action<PopPseudoServer> action,
                             [System.Runtime.CompilerServices.CallerMemberName] string callerMemberName = "",
                             [System.Runtime.CompilerServices.CallerFilePath] string sourceFilePath = "",
                             [System.Runtime.CompilerServices.CallerLineNumber] int sourceLinerNumber = 0)
    {
      Start(true, action, callerMemberName, sourceFilePath, sourceLinerNumber);
    }

    public static void Start(bool throwIfServerNotStopped,
                             Action<PopPseudoServer> action,
                             [System.Runtime.CompilerServices.CallerMemberName] string callerMemberName = "",
                             [System.Runtime.CompilerServices.CallerFilePath] string sourceFilePath = "",
                             [System.Runtime.CompilerServices.CallerLineNumber] int sourceLinerNumber = 0)
    {
      using (var server = new PopPseudoServer()) {
        server.Start(true, throwIfServerNotStopped, action, callerMemberName, sourceFilePath, sourceLinerNumber);
      }
    }

    public static void StartNotAccepting(Action<PopPseudoServer> action,
                                         [System.Runtime.CompilerServices.CallerMemberName] string callerMemberName = "",
                                         [System.Runtime.CompilerServices.CallerFilePath] string sourceFilePath = "",
                                         [System.Runtime.CompilerServices.CallerLineNumber] int sourceLinerNumber = 0)
    {
      using (var server = new PopPseudoServer()) {
        server.Start(false, false, action, callerMemberName, sourceFilePath, sourceLinerNumber);
      }
    }

    public void SetGreetingResponse(string greeting)
    {
      SetGreetingResponse(greeting, null, null, false);
    }

    public void SetGreetingResponse(string greeting,
                                    Action actionBeforeRespond,
                                    Action actionAfterRespond)
    {
      SetGreetingResponse(greeting, actionBeforeRespond, actionAfterRespond, false);
    }

    public void SetGreetingResponse(string greeting,
                                    Action actionBeforeRespond,
                                    Action actionAfterRespond,
                                    bool isQuitResponse)
    {
      if (greeting == null) {
        greetingResponse = null;
      }
      else {
        greetingResponse = new Response(null, greeting, actionBeforeRespond, actionAfterRespond);

        greetingResponse.IsQuitResponse = isQuitResponse;
      }
    }

    public void EnqueuePushResponse(string responseText)
    {
      InternalEnqueueResponse(null,
                              responseText,
                              null,
                              null,
                              false,
                              true);
    }

    public void EnqueueResponse(string responseText)
    {
      InternalEnqueueResponse(null,
                              responseText,
                              null,
                              null,
                              false,
                              false);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText)
    {
      InternalEnqueueResponse(expectedCommand,
                              responseText,
                              null,
                              null,
                              false,
                              false);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText,
                                bool isQuitResponse)
    {
      InternalEnqueueResponse(expectedCommand,
                              responseText,
                              null,
                              null,
                              isQuitResponse,
                              false);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText,
                                Action actionBeforeRespond,
                                Action actionAfterRespond)
    {
      InternalEnqueueResponse(expectedCommand,
                              responseText,
                              actionBeforeRespond,
                              actionAfterRespond,
                              false,
                              false);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText,
                                Action actionBeforeRespond,
                                Action actionAfterRespond,
                                bool isQuitResponse)
    {
      InternalEnqueueResponse(expectedCommand,
                              responseText,
                              actionBeforeRespond,
                              actionAfterRespond,
                              isQuitResponse,
                              false);
    }

    private void InternalEnqueueResponse(string expectedCommand,
                                         string responseText,
                                         Action actionBeforeRespond,
                                         Action actionAfterRespond,
                                         bool isQuitResponse,
                                         bool isPushResponse)
    {
      var response = new Response(expectedCommand,
                                  responseText,
                                  actionBeforeRespond,
                                  actionAfterRespond);

      response.IsQuitResponse = isQuitResponse;
      response.IsPushResponse = isPushResponse;

      responseQueue.Enqueue(response);
    }

    // TODO: rename
    public string DequeueRequest()
    {
      lock (commandQueue) {
        if (0 < commandQueue.Count)
          return commandQueue.Dequeue();
        else
          return null;
      }
    }

    protected override void ServerProc(TcpListener server, TcpClient client)
    {
      var commandStream = client.GetStream();

      if (greetingResponse != null) {
        Send(greetingResponse, commandStream);

        if (greetingResponse.IsQuitResponse)
          return;
      }

      var commandBuffer = new byte[0x1000];

      for (;;) {
        if (!IsConnected(client))
          return;

        Response pushResponse = null;

        lock (responseQueue) {
          if (0 < responseQueue.Count && responseQueue.Peek().IsPushResponse)
            pushResponse = responseQueue.Dequeue();
        }

        if (pushResponse != null) {
          Send(pushResponse, commandStream);

          if (pushResponse.IsQuitResponse)
            return;
          else
            continue;
        }

        /*        
         * wait for command sent
         */
        string commandText = null;

        try {
          var commandLength = commandStream.Read(commandBuffer, 0, commandBuffer.Length);

          if (commandLength == 0 && !IsConnected(client))
            return;

          commandText = OctetEncoding.EightBits.GetString(commandBuffer, 0, commandLength);

          WriteLog("C:", commandText);

          lock (commandQueue) {
            commandQueue.Enqueue(commandText);
          }
        }
        catch (IOException ex) {
          if (!client.Connected)
            return; // disconnected

          var exSocket = ex.InnerException as SocketException;

          if (exSocket != null &&
              (exSocket.SocketErrorCode == SocketError.WouldBlock || exSocket.SocketErrorCode == SocketError.TimedOut)) {
            continue;
          }

          throw;
        }

        /*
         * send response
         */
        while (responseQueue.Count == 0) {
          SetClientRequestReceived();

          if (!IsConnected(client))
            return;

          Thread.Sleep(10);
        }

        var response = responseQueue.Dequeue();

        Send(response, commandStream, commandText);

        if (response.IsQuitResponse)
          return;
      }
    }

    private readonly Queue<Response> responseQueue = new Queue<Response>();
    private readonly Queue<string> commandQueue = new Queue<string>();
    private Response greetingResponse;
  }
}



