using System;
using System.Runtime.CompilerServices;
using NUnit.Framework;

namespace Smdn.Net.Pop3.Client {
  internal class TestCases {
    public static void WithAuthenticatedClient(Action<PopPseudoServer, PopClient> testAction,
                                               [CallerMemberName] string callerMemberName = "",
                                               [CallerFilePath] string sourceFilePath = "",
                                               [CallerLineNumber] int sourceLineNumber = 0)
    {
      PopPseudoServer.Start((server) => {
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");
        server.EnqueueResponse("USER user\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("PASS pass\r\n",
                               "+OK\r\n");

        using (var client = new PopClient(new Uri(string.Format("pop://user@{0}/", server.HostPort)))) {
          client.Timeout = 5000;
          client.Profile.UsingSaslMechanisms = null;
          client.Profile.UseTlsIfAvailable = false;
          client.Profile.AllowInsecureLogin = true;

          client.Connect("pass");

          server.DequeueRequest(); // CAPA
          server.DequeueRequest(); // USER
          server.DequeueRequest(); // PASS

          testAction(server, client);
        }
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithDisconnectedClient(Action<PopPseudoServer, PopClient> testAction,
                                              [CallerMemberName] string callerMemberName = "",
                                              [CallerFilePath] string sourceFilePath = "",
                                              [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("NOOP\r\n",
                               "+OK\r\n",
                               true); // stop server

        client.KeepAlive();

        testAction(server, client);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithLoggedOutClient(Action<PopPseudoServer, PopClient> testAction,
                                           [CallerMemberName] string callerMemberName = "",
                                           [CallerFilePath] string sourceFilePath = "",
                                           [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n");

        client.Logout();

        testAction(server, client);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMessage(Action<PopPseudoServer, PopMessageInfo> action,
                                   [CallerMemberName] string callerMemberName = "",
                                   [CallerFilePath] string sourceFilePath = "",
                                   [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMessage(1L, 120L, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMessage(long messageNumber, long messageLength, Action<PopPseudoServer, PopMessageInfo> action,
                                   [CallerMemberName] string callerMemberName = "",
                                   [CallerFilePath] string sourceFilePath = "",
                                   [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               string.Format("+OK {0} {1}\r\n", messageNumber, messageLength));
        server.EnqueueResponse("LIST ",
                               string.Format("+OK {0} {1}\r\n", messageNumber, messageLength));

        var message = client.GetMessage(messageNumber);

        Assert.IsNotNull(message);
        Assert.AreEqual(messageNumber, message.MessageNumber);
        Assert.AreEqual(messageLength, message.Length);
        Assert.IsFalse(message.IsMarkedAsDeleted);

        server.DequeueRequest(); // STAT
        server.DequeueRequest(); // LIST

        action(server, message);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithDeletedMessage(Action<PopMessageInfo> action,
                                          [CallerMemberName] string callerMemberName = "",
                                          [CallerFilePath] string sourceFilePath = "",
                                          [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        server.EnqueueResponse("DELE ",
                               "+OK\r\n");

        message.MarkAsDeleted();

        server.DequeueRequest(); // DELE

        Assert.IsTrue(message.IsMarkedAsDeleted);

        server.EnqueueResponse("STAT\r\n",
                               "+OK 0 0\r\n");

        Assert.AreEqual(0L, message.Client.MessageCount);
        Assert.AreEqual(0L, message.Client.TotalSize);

        server.DequeueRequest(); // STAT

        action(message);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithClosedSessionMessage(Action<PopMessageInfo> action,
                                                [CallerMemberName] string callerMemberName = "",
                                                [CallerFilePath] string sourceFilePath = "",
                                                [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        message.Client.Disconnect();

        action(message);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }
  }
}
